use strict;
use P6C::Nodes;
use P6C::Util;

package P6C::Register;
# Gratuitous object? Not quite... It's a way of passing a temporary
# register as an unexpanded rvalue.  Useful for e.g. autoincrement,
# where we have the register lying around.

# NOTE: this would be a good place to take care of conversions between
# register types.  Right now things are always passed in P regs, but
# we could do better based on context.
use Class::Struct P6C::Register => { qw(reg $) };
sub val {
    shift->reg;
}

# IMCC code for a subroutine.
#
# XXX: the fact that e.g. P6C::prefix relies on this for argument
# information is just wrong.  This needs to be integrated with the
# stuff the parser builds up.
package P6C::IMCC::Sub;
use Class::Struct P6C::IMCC::Sub
    => { code => '$',		# function body (w/o decls and initialization)
	 localvar => '%',	# named locals
	 args => '@'		# arguments, in order passed
       };

sub code : lvalue {
    my $x = shift;
    $x->{code};
}

sub emit {
    my $x = shift;
    print <<END;
	saveall
# Parameters:
END
    foreach (@{$x->args}) {
	my ($t, $pname) = @$_;
	my $ptype = P6C::IMCC::paramtype($t);
    print <<END;
	.param $ptype	$pname
END
    }
    print "# Named locals:\n";
    while (my ($n, $t) = each %{$x->localvar()}) {
	print "\t.local $t $n\n";
    }
    # Maybe constructors for locals:
    while (my ($n, $t) = each %{$x->localvar()}) {
	next if $t == 1;	# uninitialized locals
	print "\t$n = new $t\n"
	    if P6C::IMCC::regtype($t) eq 'P';
    }
    print $x->code;
    print <<END;
	restoreall
	ret
END
}

package P6C::IMCC;
use P6C::Builtins;
use P6C::Util;
use P6C::Context;

# Map Perl types to IMCC parameter types:
my %paramtype = (int => 'int',
		  str => 'string',
		  num => 'float');
# Map Perl types to Parrot register types:
my %regtype   = (int => 'I',
		 str => 'S',
		 num => 'N');
sub regtype(*) {
    $regtype{$_[0]} || 'P';
}

sub paramtype(*) {
    $paramtype{$_[0]} || $_[0];
}

sub import {
    my @syms = qw(globalvar localvar paramvar findvar
		  add_localvar add_param
		  gentmp genlabel newtmp
		  code function
		  gen_counted_loop);
    my $class = shift;
    my $caller = caller;
    no strict 'refs';
    return unless shift eq ':all';
    foreach (@syms) {
	*{$caller . '::' . $_} = \&$_;
    }
    1;
}

our $curfunc;			# currently compiling function
our %funcs;			# all known functions
our %globals;			# global variables

sub init {			# reset state
    %funcs = ();
    %globals = ();
    undef $curfunc;
    P6C::Builtins::declare(\%funcs);
}

sub compile {			# compile input (don't emit)
    my $x = shift;
    my $ctx = new P6C::Context;	# undefined context.
    if (ref $x eq 'ARRAY') {
	# stmts.
	foreach my $stmt (@$x) {
	    $stmt->val($ctx);
	}
    } else {
	# probably single stmt.
	$x->val($ctx);
    }
}

sub function {			# switch to function, returning old one
    my $ofunc = $curfunc;
    if(@_) {
	$curfunc = $_[0];
	if ($curfunc) {
	    $funcs{$curfunc} ||= new P6C::IMCC::Sub;
	}
    }
    return $ofunc;
}

sub code {			# add code to current function
    die "Code must live within a function" unless defined $curfunc;
    $funcs{$curfunc}->code .= join "\n", @_;
}

sub emit {			# emit all code
    die "Must define main" unless $funcs{main};
    print <<'END';
.sub _main
	call main
	end
	ret
END
    while (my ($name, $sub) = each %funcs) {
	unless($sub->code) {
	    print STDERR "Skipping empty sub $name (builtin or external?)\n";
	    next;
	}
	print ".sub $name\n";
	$sub->emit;
    }
    P6C::Builtins::emit;
}

##############################
# Name lookup
sub globalvar($) {		# find a named global
    # Globals always exist...
    # except when they don't
    unimp 'No globals yet';
    my $name = shift->mangled_name;
    $globals{$name} ||= 1;
    $name;
}

sub localvar($) {		# find a named local
    # Locals must be explicitly declared
    die "Local variable outside function" unless defined $curfunc;
    my $name = shift->mangled_name;
    if (exists $funcs{$curfunc}->localvar->{$name}) {
	$name;
    } else {
	undef;
    }
}

sub add_localvar($) {		# add a named local (with optional initializer)
    my ($var) = @_;
    if (localvar($var)) {
	warn 'Redeclaring lexical '.$var->name." in $curfunc\n";
    }
    my $name = $var->mangled_name;
    $funcs{$curfunc}->localvar->{$name} ||= $var->type;
    $name;
}

sub paramvar($) {
    my $var = shift->mangled_name;
    return $var if grep {
	$_->[1] eq $var
    } @{$funcs{$curfunc}->args};
    return undef;
}

# ($name, $isglobal) = findvar($name, $inittype)
# $name = findvar($name, $inittype)
#	find a variable, or create it with an initial type.
sub findvar($) {
    my ($var) = @_;
    my $name;
    $name = paramvar($var) || localvar($var) || globalvar($var);
    return wantarray ? ($name, 0) : $name;
}

sub add_param($) {		# Add a parameter to current function
    my ($p) = @_;
    push @{$funcs{$curfunc}->args()},
	[$p->var->type, $p->var->mangled_name];
}

##############################
# Generate temporary names:
my $lastsym = 0;
my $lasttmp = 0;
my $lastlabel = 0;
sub gensym(*) {
    $_[0] . ++$lastsym
} 

sub _gentmp(*) {		# uninitialized temporary (internal)
    '$' . $_[0] . ++$lasttmp
}

sub gentmp(;*) {			# uninitialized temporary
    _gentmp(regtype($_[0] || 'PerlUndef'));
}

sub genlabel(;$) {		# new label (optionally annotated)
    'L_'.$_[0]. ++$lastlabel
}

sub newtmp(;*) {			# initialized temporary
    my $type = shift || 'PerlUndef';
    my $reg = regtype $type;
    my $name;
    if ($reg eq 'S') {
 	$name = _gentmp S;
	code(<<END);
	$name = ""
END
    } elsif ($reg eq 'I' || $reg eq 'N') {
	$name = _gentmp $reg;
	code(<<END);
	$name = 0
END
    } else {
	$name = _gentmp P;
	die unless $type;
	use Carp;
	code(<<END);
	$name = new $type
END
    }
    return $name;
}

#  gen_counted_loop($counter, $body) -- generate a counted loop using
#  $counter as the repetition count.  The loop will iterate over
#  values between 0 and $counter - 1, inclusive.
sub gen_counted_loop {
    my ($count, $body) = @_;
    my $start = genlabel;
    my $end = genlabel;
    return <<END;
$start:	if $count == 0 goto $end
	dec $count
	$body
	goto $start
$end:
END
}

package P6C::Binop;
use P6C::IMCC ':all';
use P6C::Util;
use P6C::Context;

sub simple_binary {
    my ($op, $l, $r) = @_;
    my $ltmp = $l->val;
    my $rtmp = $r->val;
    my $dest = newtmp;
    code("\t$dest = $ltmp $op $rtmp\n");
    return $dest;
}

sub slow_pow {
    my ($l, $r) = @_;
    my $dest = newtmp;
    my $lv = $l->val;
    my $rv = $r->val;
    my $cnt = gentmp 'int';
    code(<<END);
# POW
	$dest = 1
	$cnt = $rv
END
    code(gen_counted_loop($cnt, "$dest = $dest * $lv\n"));
    return $dest;
}

sub do_assign {
    my $l = shift;
    return $l->assign(@_);
}

sub do_logand {
    my ($l, $r) = @_;
    my $dest = newtmp;
    my $thenlab = genlabel 'logical_and';
    my $endlab = genlabel 'logical_and';
    my $res = $l->val;
    code(<<END);
	if $res goto $thenlab
	goto $endlab
$thenlab:
END
    $res = $r->val;
    code(<<END);
	$dest = $res
$endlab:
END
    return $dest;		# will be undef if first failed.
}

sub do_logor {
    my ($l, $r) = @_;
    my $dest = newtmp;
    my $endlab = genlabel 'logical_or';
    my $res = $l->val;
    code(<<END);
# LOGICAL OR
	$dest = $res
	if $dest goto $endlab
END
    $res = $r->val;
    code(<<END);
	$dest = $res
# END_LOGICAL_OR
$endlab:
END
    return $dest;
}

sub do_defined {
    my ($l, $r) = @_;
    my $val = $l->val;
    my $itmp = gentmp 'int';
    my $res = newtmp;
    my $endlab = genlabel 'defined';
    code(<<END);
	$res = $val
	$itmp = defined $res
	if $itmp goto $endlab
END
    $val = $r->val;
    code(<<END);
	$res = $val
$endlab:
END
    return $res;
}

sub do_concat {
    # XXX: PMC concat doesn't seem to work, so we have to go through strings.
    my ($l, $r) = @_;
    my $lt = gentmp 'str';
    my $rt = gentmp 'str';
    my $restmp = gentmp 'str';
    my $res = newtmp;
    my $lval = $l->val;
    my $rval = $r->val;
    code(<<END);
	$lt = $lval
	$rt = $rval
	$restmp = $lt . $rt
	$res = $restmp
END
    return $res;
}

sub do_array {
    my ($l, $r, $ctx) = @_;
    use Carp 'cluck';
    cluck "Should provide context to comma operator" unless $ctx;
    my @things = flatten_leftop($l, ',');
    push @things, $r;
    if (!$ctx || $ctx->is_scalar) {
# 	print STDERR "list in scalar context\n";
	for (@things[0..$#things - 1]) {
	    $_->val;
	}
	return $things[-1]->val;
    } elsif ($ctx->is_tuple) {
# 	print STDERR "list in tuple context\n";
	my @ret;
	my $min = @things < $ctx->nelem ? @things : $ctx->nelem;
	for my $i (0..$min - 1) {
	    $ret[$i] = $things[$i]->val($ctx->type($i));
	}
	for my $i ($min .. $#things) {
	    $things[$i]->val;
	}
	return [@ret];
    } elsif ($ctx->is_array) {
# 	print STDERR "list in array context\n";
	my $tmp = newtmp 'PerlArray';
	code("\t$tmp = ".@things."\n");
	# XXX: kludge until the assembler can handle "Px[Iy] = Pz"
	my $stmp = gentmp 'str';
	for my $i (0..$#things) {
	    my $item = $things[$i]->val;
	    code(<<END);
	$stmp = $item
	$tmp\[$i] = $stmp
END
	}
	return $tmp;
    } else {
	use Data::Dumper;
	unimp "Can't handle context ".Dumper($ctx);
    }
}

sub do_repeat {
    unimp 'repeat';
}

my %ops =
(
 '+' => sub { simple_binary('+', @_) },
 '-' => sub { simple_binary('-', @_) },
 '*' => sub { simple_binary('*', @_) },
 '/' => sub { simple_binary('/', @_) },
 '%' => sub { simple_binary('%', @_) },
 '>>' => sub { simple_binary('>>', @_) },
 '<<' => sub { simple_binary('<<', @_) },
 '|' => sub { simple_binary('|', @_) },
 '&' => sub { simple_binary('&', @_) },
 '~' => sub { simple_binary('~', @_) },
# '_' => sub { simple_binary('.', @_, 'PerlString') },
 '_' => \&do_concat,
 '**' => \&slow_pow,
 '=' => \&do_assign,
 '||' => \&do_logor,
 '&&' => \&do_logand,
 '~~' => sub { simple_binary('~~', @_) },
 '//' => \&do_defined,
 ',' => \&do_array,
 'x' => \&do_repeat,
);

sub val {
    my $x = shift;
    if (ref $x->op eq 'P6C::hype') {
	return do_hyped($x->op->op, $x->l, $x->r, @_);
    }
    if ($ops{$x->op}) {
	return $ops{$x->op}->($x->l, $x->r, @_);
    } elsif($x->op() =~ /^([^=]+)(=)$/
	    && $ops{$1}) {
	return $ops{'='}->($x->l,
			   new P6C::Binop(op => $1, l => $x->l, r => $x->r),
			   @_);
    } else {
	use Data::Dumper;
	confess "No such operator `".$x->op()."'".Dumper($x);
    }
}

# XXX: To support the indexing hack, we need to know what kind of
# temporaries to use...

my %optype;
BEGIN {
    my %opmap = (int => [ qw(>> << | & ~ ~~)],
		 num => [ qw(+ - * / % **)]);
    while (my ($t, $ops) = each %opmap) {
	@optype{@$ops} = ($t) x @$ops;
    }
}

sub simple_hyped {
    my ($op, $targ, $lindex, $rindex) = @_;
    my $optype = $optype{$op} or unimp "Can't hype $op yet";
    my $ltmp = gentmp $optype;
    my $rtmp = gentmp $optype;
    my $dest = gentmp $optype;
    return <<END;
	$ltmp = $lindex
	$rtmp = $rindex
	$dest = $ltmp $op $rtmp
	$targ = $dest
END
}

sub isarray {			# XXX: hack to do context.
    my $x = shift;
    return ($x->isa('P6C::Binop') && ref($x->op) eq 'P6C::hype')
	|| ($x->isa('P6C::variable') && $x->type eq 'PerlArray');
}

sub do_hyped {
    my ($op, $l, $r) = @_;
    if (isarray($l) && isarray($r)) {
	return hype_array_array(@_);
    } elsif (isarray($l)) {
	return hype_array_scalar(@_);
    } elsif (isarray($r)) {
	return hype_scalar_array(@_);
    } else {
	diag "Tried to hyper-operate two scalars";
	return simple_binary(@_);
    }
}

sub hype_array_scalar {
    my ($op, $l, $r) = @_;
    my $lval = $l->val;
    my $rval = $r->val;
    my $len = gentmp 'int';
    my $dest = newtmp 'PerlArray';
    code(<<END);
	$len = $lval
	$dest = $len
END
    my $code = simple_hyped($op, "$dest\[$len]", "$lval\[$len]", $rval);
    code(gen_counted_loop($len, $code));
    return $dest;
}

sub hype_scalar_array {
    my ($op, $l, $r) = @_;
    my $lval = $l->val;
    my $rval = $r->val;
    my $len = gentmp 'int';
    my $dest = newtmp 'PerlArray';
    code(<<END);
	$len = $rval
	$dest = $len
END
    my $code = simple_hyped($op, "$dest\[$len]", $lval, "$rval\[$len]");
    code(gen_counted_loop($len, $code));
    return $dest;
}

sub hype_array_array {
    my ($op, $l, $r) = @_;
    my $lval = $l->val;
    my $rval = $r->val;
    my $llen = gentmp 'int';
    my $rlen = gentmp 'int';
    # XXX: gross indexing hack again:

    my $cntlabel = genlabel;
    my $dest = newtmp 'PerlArray';
    my $looptop = genlabel 'hyper';
    my $loopend = genlabel 'hyper_end';
    code(<<END);
	$llen = $lval
	$rlen = $rval
	if $llen > $rlen goto $cntlabel
	$llen = $rlen
$cntlabel:
	$dest = $llen
END
    my $code
	= simple_hyped($op, "$dest\[$rlen]", "$lval\[$rlen]", "$rval\[$rlen]");
    code(gen_counted_loop($rlen, $code));
    return $dest;
}

sub P6C::sv_literal::val {
    my $x = shift;
    my $tmp = gentmp PerlUndef;
    my $type = $x->type;
    # XXX: these are actually _references_.  But we don't support them
    # anyways.
    die "Don't support ".$type if $type =~ /Perl(Hash|Array)/;
    my $val = $x->lval;
    code(<<END);
	$tmp = new $type
	$tmp = $val
END
    return $tmp;
}

package P6C::prefix;
use P6C::IMCC ':all';
use P6C::Util;
use P6C::Context;

sub make_noarg {
    my $block = shift;
    # XXX: really, just make it a no-arg sub before we inline it.
    $block->params(new P6C::params req => [], opt => [], rest => undef);
}

sub prefix_if {
    my $x = shift;
    my $end = genlabel "endif";
    my $tmp = newtmp;
    my $nextlab;
    foreach (@{$x->args}) {
	my ($sense, $test, $block) = @$_;
	$sense ||= $x->name;
	if ($nextlab) {
	    code("$nextlab:\n");
	}
	$nextlab = genlabel 'if';
	if (!ref $test) {
	    make_noarg($block);
	    $block->val;
	} else {
	    my $v = $test->val;
	    if ($sense =~ /if$/) { # (els)?if
		code(<<END);
	$tmp = ! $v
	if $tmp goto $nextlab
END
	    } else {		# (els)?unless
		code(<<END);
	if $v goto $nextlab
END
	    }
	    make_noarg($block);
	    $block->val;
	    code(<<END);
	goto $end
END
	}
    }
    code(<<END);
$nextlab:
$end:		# END OF @{[$x->name]}
END
    return undef;
}

sub gen_sub_call {
    my ($x, $func) = @_;
    # gather args.
    my @argtypes = map { $_->[0] } @{$func->args};
    my $argctx = new P6C::Context type => \@argtypes;
    my $args = $x->args->val($argctx);
    if (ref($args) ne 'ARRAY') {
	$args = [$args];
    }
    if (@$args != @{$func->args}) {
	die "Wrong number of arguments for ".$x->name.": got ".@$args
	    .", expected ".@{$func->args};
    }
    # XXX: Not sure this is the right place to reverse...
    foreach (reverse @$args) {
	code("\t.arg	$_\n");
    }
    my $name = $x->name;
    code("\tcall	$name\n");
    return newtmp;
}

sub prefix_neg {
    my $x = shift;
    my $tmp = $x->args->val;
    my $res = newtmp;
    code(<<END);
	$res = - $tmp
END
    return $res;
}

my %prefix_ops =
(
 'if' => \&prefix_if,
 'unless' => \&prefix_if,
 '-' => \&prefix_neg,
);

sub P6C::prefix::val {
    my $x = shift;
    # XXX: temporary hack.
    if (exists $P6C::IMCC::funcs{$x->name}) {
	return gen_sub_call($x, $P6C::IMCC::funcs{$x->name});
    } elsif (exists $prefix_ops{$x->name}) {
	return $prefix_ops{$x->name}->($x);
    } else {
	unimp "Prefix operator ".$x->name();
    }
}

package P6C::compare;
use P6C::IMCC ':all';

our %type;
BEGIN {
    $type{$_} = 'num' for qw(<= == >= < > !=);
    $type{$_} = 'str' for qw(eq ne ge le lt gt);
}

my %imccop;
BEGIN {
    @imccop{qw(<= == >= < > !=)} = qw(<= == >= < > !=);
    @imccop{qw(le eq ge lt gt ne)} = qw(<= == >= < > !=);
}

sub gen_compare {
    my ($op, $l, $r, $fail) = @_;
    my $label = genlabel;
    $op = $imccop{$op};
    code(<<END);
	if $l $op $r goto $label
	goto $fail
$label:
END
}
our %ops;
BEGIN {
    for my $op (qw(<= == >= < > != le eq ge lt gt ne)) {
	$ops{$op} = \&gen_compare;
    }
}

sub val {
    my $self = shift;
    die "Nothing to compare" if $self->size == 0;
    my $x = $self->seq;
    my $result = gentmp 'int';
    my %tmps;
    $tmps{0} = newtmp;
    $tmps{2} = newtmp;
    my $res = newtmp;
    my $lasttype;
    my $fail = genlabel 'comparison';
    code("\t$result = 0\n");
    $tmps{0} = $x->[0]->val;
    for (my $i = 1; $i < $#{$x}; $i += 2) {
	my $op = $x->[$i];
	my $type = $type{$op} or die "No such op: $op";
	unless ($tmps{"$type 0"}) {
	    # Initialize temporaries for this type if we haven't already.
	    $tmps{"$type 0"} = gentmp $type;
	    $tmps{"$type 2"} = gentmp $type;
	}

	# Only recompute expression values if necessary
	my $ltmp = $tmps{$type.' '.(($i - 1) % 4)};
	my $rtmp = $tmps{$type.' '.(($i + 1) % 4)};
	if ($lasttype ne $type) {
	    code(<<END);
	$ltmp = $tmps{($i - 1) % 4}
END
	}
	$lasttype = $type;
	$tmps{(($i + 1) % 4)} = $x->[$i + 1]->val;
	code(<<END);
	$rtmp = $tmps{(($i + 1) % 4)}
END
	$ops{$op}->($op, $ltmp, $rtmp, $fail);
    }
    code(<<END);
	$result = 1
$fail:
    $res = $result
END
    return $res;
}

sub P6C::sub_def::val {
    use P6C::IMCC ':all';
    use P6C::Util;
    my $x = shift;

    my $ofunc = function($x->name);
    $x->closure->val;
    function($ofunc);
}

package P6C::closure;
use P6C::Util;
use P6C::IMCC ':all';

sub val {
    my $x = shift;
    if (!$x->params || $x->params->min != $x->params->max) {
	unimp "No arrays, so we need fixed param lists",
	    ($x->params ? $x->params->min() . ' - ' . $x->params->max
	     : '0 - any');
    }
    foreach my $p (@{$x->params->req}) {
	add_param($p);
    }
    foreach my $stmt (@{$x->block}) {
	$stmt->val;
    }
}

package P6C::variable;
use P6C::IMCC ':all';
use P6C::Context;
##############################
# Name mangling
my %mangle = (qw(! __BANG
		 _ __UNDERSCORE));

my %mangle_sig = (qw($ SV_
		     @ AV_
		     % HV_
		     & CV_));

sub mangled_name {
    my $x = shift;
    my $name = $mangle{$x->name} || $x->name;
    $name =~ s/^(.)/$mangle_sig{$1} || $1/e;
    return $name;
}

sub val {
    return scalar findvar(shift);
}

sub assign {
    my ($x, $thing) = @_;
    my $name = findvar($x);
    my $ctx = new P6C::Context type => $x->type;
    my $tmpv = $thing->val($ctx);
    use Data::Dumper;
    code("\t$name = $tmpv\n");
    return $name;
}

package P6C::ternary;
use P6C::IMCC ':all';

sub val {
    my $x = shift;
    my $tmp = newtmp;
    my ($thenlab, $endlab) = (genlabel("ternary_then"),
			      genlabel("ternary_end"));
    code(<<END);
# START TERNARY
END
    my $ifval = $x->if->val;
    code(<<END);
	if $ifval goto $thenlab
END
    my $elseval = $x->else->val(@_);
    code(<<END);
	$tmp = $elseval
	goto $endlab
$thenlab:
END
    my $thenval = $x->then->val(@_);
    code(<<END);
	$tmp = $thenval
$endlab:
# END TERNARY
END
    return $tmp;
}

sub assign {
    my ($x, $thing) = @_;
    my $tmp = newtmp;
    my ($thenlab, $endlab) = (genlabel("ternary_then"),
			      genlabel("ternary_end"));
    code(<<END);
# START TERNARY
END
    my $ifval = $x->if->val;
    code(<<END);
	if $ifval goto $thenlab
END
    my $elseval = $x->else->assign($thing);
    code(<<END);
	$tmp = $elseval
	goto $endlab
$thenlab:
END
    my $thenval = $x->then->assign($thing);
    code(<<END);
	$tmp = $thenval
$endlab:
# END TERNARY
END
    return $tmp;
}

package P6C::decl;
use P6C::IMCC ':all';
use P6C::Util;
use P6C::Context;

sub val {
    my $x = shift;
    if ($x->qual && $x->qual->scope ne 'my') {
	unimp 'global variables';
    }
    if (@{$x->props}) {
	unimp 'variable properties';
    }
    add_localvar($_) for @{$x->vars};
}

sub assign {
    my ($x, $thing) = @_;

    my @ctx = map { $_->type } @{$x->vars};
    my $ctx = new P6C::Context type => (@ctx == 1 ? $ctx[0] : \@ctx);
    use Data::Dumper;
    my $tmpv = $thing->val($ctx);
    if (@ctx == 1) {
	if (ref($tmpv) eq 'ARRAY') {
	    $tmpv = $tmpv->[-1];
	}
	add_localvar($x->vars(0));
	$x->vars(0)->assign(new P6C::Register reg => $tmpv);
	return undef;
    } else {
	if (ref $tmpv ne 'ARRAY') {
	    $tmpv = [$tmpv];
	}
	if (@ctx != @$tmpv) {
	    diag 'Different numbers of values in array initialization';
	}
	my $min = @ctx < @$tmpv ? @ctx : @$tmpv;
	for my $i (0 .. $min - 1) {
	    add_localvar($x->vars($i));
	    $x->vars($i)->assign(new P6C::Register reg => $tmpv->[$i]);
	}
	for my $i ($min .. $#ctx) {
	    add_localvar($x->vars($i));
	}
	return undef;
    }
}

package P6C::incr;
use P6C::IMCC ':all';

my %inplace_op = ('++' => 'inc', '--' => 'dec');
my %outaplace_op = ('++' => '+ 1', '--' => '- 1');

sub val {
    my $x = shift;
    my $ret;
    if ($x->post) {
	my $op = $outaplace_op{$x->op}
	    or die $x->op().' increment not understood';
	$ret = $x->thing->val;
	my $tmp = newtmp;
	code("\t$tmp = $ret $op\n"); # yep, the order's right.
	$x->thing->assign(new P6C::Register reg => $tmp);
	return $ret;
    } else {
	my $op = $inplace_op{$x->op}
	    or die $x->op().' increment not understood';
	$ret = $x->thing->val;
	code("\t$op $ret\n");
	$x->thing->assign(new P6C::Register reg => $ret);
	return $ret;
    }
}

package P6C::subscript_exp;
use P6C::Util;
use P6C::IMCC ':all';

sub subscript_indices {
    my $subsc = shift;
    if ($subsc->type ne 'PerlArray') {
	unimp "Only array subscripts, not ".$subsc->type;
    }
    my @indices;
    if ($subsc->indices->isa('P6C::Binop')) {
	@indices = $subsc->indices->flatten_leftop(',');
    } else {
	push @indices, $subsc->indices;
    }
    return @indices;
}

sub val {
    my $x = shift;
    if (@{$x->subscripts} > 1) {
	unimp "multi-level subscripting";
    }
    my @indices = subscript_indices($x->subscripts(0));
    my $thing = $x->thing->val;
    my $ret;
    if (@indices > 1) {
	$ret = newtmp 'PerlArray';
	my $tmp = gentmp 'str';	# XXX: should be PMC, but for now...
	my $itmp = gentmp 'int';
	for my $i (0..$#indices) {
	    my $idx = $indices[$i]->val;
	    code(<<END);
	$itmp = $idx
	$tmp = $thing\[$itmp]
	$ret\[$i] = $tmp
END
	}
    } else {
	$ret = newtmp;
	my $stmp = gentmp 'str'; # XXX: once again, should be PMC
	my $itmp = gentmp 'int';
	my $idx = $indices[0]->val;
	code(<<END);
	$itmp = $idx
	$stmp = $thing\[$itmp]
	$ret = $stmp
END
    }
    return $ret;
}

sub assign {
    my ($x, $thing) = @_;
    if (@{$x->subscripts} > 1) {
	unimp "multi-level subscripting";
    }
    unimp 'Assignment to array-slice';
}

1;
