package P6C::Parser;
use Parse::RecDescent;
use strict;

use vars qw(%WANT %CLASSES $FUNCTION_ARGS $grammar);

######################################################################
# Argument context for functions and control structures
######################################################################

%WANT = ();
%CLASSES = ();

sub Precompile {
    shift;
    return Parse::RecDescent->Precompile($grammar, @_);
}

sub new {
    shift;
    return Parse::RecDescent->new($grammar, @_);
}

sub add_function {
    my $fname = shift->[1];
    $WANT{$fname} = shift || $FUNCTION_ARGS;
    1;
}

sub add_class {		# seen class.
    my $c = shift->[1];
    $CLASSES{$c} = $c;
    1;
}

##############################
# Functions (list operators):
# XXX: many of these need their own special want_* rules

INIT {
$FUNCTION_ARGS = 'maybe_comma';

my @builtin_funcs = qw(crypt index pack rindex sprintf substr
		       join unpack split
		       push unshift splice
		       warn die print printf read select syscall sysread
		       seek sysseek syswrite truncate write
		       vec
		       chmod chown fcntl ioctl link open opendir
		       rename symlink sysopen unlink
		       return fail
		       puts
		       not);
@WANT{@builtin_funcs} = ($FUNCTION_ARGS) x @builtin_funcs;

##############################
# Loop control
my @loop_control = qw(redo last next continue);
@WANT{@loop_control} = ('maybe_label') x @loop_control;

##############################
# Unary operators
# XXX: need to handle default $_
my @unary_ops = qw(chop chomp chr hex lc lcfirst length
		   ord reverse uc ucfirst
		   abs atan2 cos exp hex int log oct rand sin sqrt srand
		   pop shift
		   delete each exists keys values
		   defined undef
		   chdir chroot glob mkdir rmdir stat umask
		   close);
@WANT{@unary_ops} = ('prefix') x @unary_ops;

##############################
# Control operators
my @control = qw(for foreach while given when default sort grep map);
@WANT{@control} = map { "want_for_$_" } @control;
$WANT{if} = $WANT{unless} = 'want_for_if';

##############################
# Named blocks
my @special_blocks = qw(CATCH BEGIN END INIT AUTOLOAD
			PRE POST NEXT LAST FIRST
			try do);
@WANT{@special_blocks} = ('closure') x @special_blocks;

##############################
# Classes (builtin and otherwise)
my @builtin_types = qw(int real str HASH ARRAY SCALAR
		       true false); # XXX: these are really constants
@CLASSES{@builtin_types} = @builtin_types;

# Handle comments:
BEGIN {
    $Parse::RecDescent::skip = '(\s*(#[^\n]*?\n)?)+';
    $::RD_AUTOACTION = q { bless [@item], 'P6C::'.$item[0] };
}

$grammar = <<'ENDSUPPORT';
#!perl
######################################################################
# Support functions and variables
{
    use vars '$no_comma';
    use vars qw(%KEYWORDS %CLASSES %WANT);
    # Regexen used in the parser:
    use vars qw($NAMEPART $COMPARE $CONTEXT $MULDIV $PREFIX $ADDSUB
		$LOG_OR $LOGOR $FILETEST $ASSIGN);
    
BEGIN {
    no strict 'refs';
    for (qw(add_class add_function CLASSES WANT)) {
	*$_ = *{'P6C::Parser::'.$_};
    }
}

BEGIN {
    $NAMEPART	= qr/[a-zA-Z_][\w_]*/;
    $COMPARE	= qr{cmp | eq | [gnl]e | [gl]t
		     | <=> | [<>=!]= | < | > }x;
    $CONTEXT	= qr{[\%\@\$\&*_?] | \+(?!\+)}x;
    $MULDIV	= qr{[\%*x] | /(?!/)}x;
    $PREFIX	= qr{[!~\\] | -(?![->])}x;
    $ADDSUB	= qr{[-+_]};
    $LOG_OR	= qr{x?or | err}x;
    $LOGOR	= qr{\|\| | ~~ | //}x;
    $FILETEST	= qr{-[rwxoRWXOezsfdlpSbctugkTBMAC]+};
    $ASSIGN	= qr{(?:
		     ! | :	# != and :=
		     | //
		     | &&? | \|\|? | ~~?
		     | << | >>
		     | $ADDSUB
		     | $MULDIV
		     | \*\*
		    )? =}x;
}

# HACK to distinguish between "my ($a, $b) ..." and "foo ($a, $b)".
# Don't need all keywords in here, but only the ones that cause
# problems.
%KEYWORDS = ();

BEGIN {
    @KEYWORDS{qw(my our temp
	         if unless until while for
	         and or xor err)} = 1;
}

# (see Parse::RecDescent::Consumer)
sub consumer {
    my $t = shift;
    my $old_len = length $t;
    return sub {
	my $len = length($_[0]);
	return substr($t, 0, ($old_len - $len));
    };
}

my %since;

sub check_end {
    my ($type, $text) = @_;
    if ($since{$type}) {
	local $_ = $since{$type}->($text);
	return m/\A[\s\n]+\z/ || undef;
    }
    return undef;
}

sub mark_end {
    my ($type, $text) = @_;
    $since{$type} = consumer($text);
}

}

ENDSUPPORT

$grammar .= <<'ENDGRAMMAR';
##############################
# Literals:

sv_literal:	  /(?:\d+(?:\.\d+)?|\.\d+)(?:[Ee]-?\d+)?/
		| '{' <commit> hv_seq '}'
		| '[' <commit> av_seq(?) ']'
		| <perl_quotelike>

av_seq:		  semi /[;,]?/

hv_seq:		  <leftop: pair ',' pair> /,?/

##############################
# Variables:
variable:	  sigil <skip:''> varname

sigil:		  /[\@\%\$\&]/

varname:	  name
		| /\d+/
		| /[\!_]/
		| '^' <commit> namepart
		| ('*')(?) '{' <skip:'\s*'> (scalar_expr | name) '}'

name:		  /(?:::|\.|\*)?$NAMEPART(::$NAMEPART)*/o

namepart:	  /$NAMEPART/o

maybe_namepart:	  namepart |

##############################
# Expressions

hype:		  '^' <commit> <skip:''> <matchrule:$arg[0]>
		| <matchrule:$arg[0]>

hv_indices:	  /[\w_]+(?=\s*})/ | comma

arglist:	  '(' comma(?) ')'

subscript:	  <skip:''> '{' <commit> <skip:$item[1]> hv_indices '}'
		| <skip:''> '[' <commit> <skip:$item[1]> av_seq ']'
		| '(' comma(?) ')'

subscriptable:	  name <commit>
			{ exists($KEYWORDS{$item[1][1]}) ? undef : 1 }
			arglist
		| '.' <commit> <skip:''> namepart(?)
		| '(' <commit> av_seq(?) ')'
		| variable

context:	  /$CONTEXT/o
# context:	  '%' | '@' | '$' | '&' | '*' | '_' | '?'
# 		| /\+(?!\+)/ # numeric context...

term:		  '<' <commit> expr(?) '>'
		| subscriptable <commit> subscript(s?)
		| /$CONTEXT/o <commit> term
		| sv_literal
		| class
		| closure

apply_rhs:	  namepart <commit> subscript(s?)
		| subscript(s)

apply:		  <leftop: term hype['apply_op'] apply_rhs>
apply_op:	  '.'

incr:		  hype['incr_op'] <commit> apply
		| apply hype['incr_op'](?)
incr_op:	  '++' | '--'

pow:		  <leftop: incr hype['pow_op'] prefix>
pow_op:		  '**'

prefix:		  filetest_op <commit> prefix
		| hype['prefix_op'] <commit> prefix
		| name { $WANT{$item[1][1]} } <matchrule:$item[2]>
		| pow

# prefix_op:	  '!' | '~' | '\\' | /-(?![->])/
prefix_op:	  /$PREFIX/o
filetest_op:	  /$FILETEST/o

pair:		  namepart '=>' <commit> prefix
		| prefix '=>' prefix
maybe_pair:	  namepart '=>' <commit> prefix
		| prefix ('=>' prefix)(?)

match:		  <leftop: maybe_pair hype['match_op'] maybe_pair>
match_op:	  '=~' | '!~'

muldiv:		  <leftop: match hype['muldiv_op'] match>
# muldiv_op:	  '*' | '/' | '%' | 'x'
muldiv_op:	  /$MULDIV/o

addsub:		  <leftop: muldiv hype['addsub_op'] muldiv>
# addsub_op:	  '+' | '-' | '_'
addsub_op:	  /$ADDSUB/o

bitshift:	  <leftop: addsub hype['bitshift_op'] addsub>
bitshift_op:	  '<<' | '>>'

compare:	  <leftop: bitshift hype['compare_op'] bitshift>
compare_op:	  /$COMPARE/o
# compare_op:	  '<=>' | '<=' | '==' | '>=' | '<' | '>' | '!='
# 		| 'eq' | 'ge' | 'ne' | 'le' | 'lt' | 'gt' | 'cmp'

bitand:		  <leftop: compare hype['bitand_op'] compare>
bitand_op:	  /&(?!&)/

bitor:		  <leftop: bitand hype['bitor_op'] bitand>
bitor_op:	  /([|~])(?!\1)/

logand:		  <leftop: bitor hype['logand_op'] bitor>
logand_op:	  '&&'

logor:		  <leftop: logand hype['logor_op'] logand>
logor_op:	  '||' | '~~' | '//'
# logor_op:	  /$LOGOR/o

range:		  logor (range_op logor)(?)
range_op:	  '..'

ternary:	  range ('??' ternary '::' ternary)(?)

scope:		  'my' | 'temp' | 'our'
class:		  name { $CLASSES{$item[1][1]} }
			{ bless ['class', $item[1]], 'P6C::class' }

scope_class:	  scope <commit> class(?)
		| class

property:	  name { ($item[1][1] ne $arg[0]) || undef }
			arglist(?)

and_prop:	  "$arg[0]" <commit> property[$arg[0]]
		| property[$arg[0]]
props:		  "$arg[0]" <commit> property and_prop[$arg[0]](s?)
		| # nothing
decl:		  '(' <commit> <leftop: variable ',' variable> ')' props['are']
		|  variable props['is']

assign:		  assign_lhs assign_rhs(s?)

assign_lhs:	  scope_class decl
		| ternary
assign_rhs:	  hype['assign_op'] scalar_expr

assign_op:	  /$ASSIGN/o
# assign_op:	  /[!:]?=/ <commit>
# 		| assignable_op <skip:''> '='
# assignable_op:	  '//'
# 		| logand_op | logor_op
# 		| bitand_op | bitor_op | bitshift_op
# 		| addsub_op | muldiv_op | pow_op

scalar_expr:	  assign but(s?)
but:		  'but' assign

comma:		  <leftop: <matchrule:@{[$arg[0] || 'scalar_expr']}>
			comma_op <matchrule:@{[$arg[0] || 'scalar_expr']}> >
comma_op:	  ','
maybe_comma:	  ...!'(' comma[$arg[0]]
		| ...!'('

semi:		  <leftop: expr semi_op expr>
semi_op:	  ';'

adverb:		  comma adv_clause(?)
adv_clause:	  /:(?!:)/ comma['scalar_expr']

log_AND:	  <leftop: adverb hype['log_AND_op'] adverb>
log_AND_op:	  'and'

log_OR:		  <leftop: log_AND hype['log_OR_op'] log_AND>
# log_OR_op:	  'or' | 'xor' | 'err'
log_OR_op:	  /$LOG_OR/o

expr:		  log_OR

##############################
# Parameters
params:		   <rulevar: local $no_comma>
params:           '(' <commit> { $no_comma = 1 }
			maybe_params opt_params rest_param ')'
                | # nothing

maybe_params:     _params { $no_comma = undef;
			    bless([@item[0,1]], 'P6C::maybe_params') }
                | # nothing

opt_params:       ';' <commit> { $no_comma = 1 } maybe_params
                | # nothing

rest_param:       { $no_comma } '*' <commit> '@' namepart
                | ',' <commit> '*' '@' namepart
                | # nothing

_params:	  <leftop: param ',' param>
param:		  scope_class(?) variable props['is'] initializer(?)
initializer:	  hype['assign_op'] scalar_expr

##############################
# Statements

prog:		  /\A/ stmts /\z/
		| <error>

stmts:	  	  <leftop: stmt stmt_sep stmt> stmt_sep(?)
		| # nothing

stmt_sep:	  ';'
		| { check_end('block', $text) }
		| { check_end('label', $text) }

stmt:		  label ':' { mark_end('label', $text);1 } ''
		| directive <commit> name comma(?)
		| 'method' <commit> name params props['is'] block
		| 'loop' <commit>
			'(' scalar_expr(?)
			';' scalar_expr(?)
			';' scalar_expr(?) ')'
			block
		| scope(?) 'sub' <commit> name { add_function($item[4]);1 }
			params props['is'] block
		| scope(?) 'class' <commit> name { add_class($item[4]);1 }
			props['is'] block
		| expr guard(?)

directive:	  /package|module|use/
guard:		  { check_end('block', $text) ? undef : 1 } _guard
_guard:		  /if|unless|while|until/ <commit> scalar_expr
		| 'for' expr

maybe_label:	  label(?)
label:		  namepart { exists($KEYWORDS{$item[1][1]}) ? undef
				: bless(['label', $item[1]], 'P6C::label') }

block:		  start_block '...' <commit> '}'
			{ mark_end('block', $text);1; } ''
		| start_block stmts '}'
			{ mark_end('block', $text);1; } ''

start_block:	  <skip:''> /\s*(?<![^\n\s]){\s*/m

closure:	  '->' '(' <commit> _closure_args(?) ')' block
		| '->' <commit> _closure_args(?) block
		| block

_closure_args:	  <leftop: comma['variable'] ';' comma['variable']>

want_for_for:	  av_seq closure
want_for_foreach: maybe_decl '(' expr ')' block
want_for_given:	  scalar_expr closure
want_for_while:	  scalar_expr closure
want_for_when:	  comma closure
want_for_default: closure

want_for_if:	  scalar_expr closure elsif(s?) else(?)
elsif:		  /els(?:if|unless)/ scalar_expr closure
else:		  'else' closure

want_for_grep:	  scalar_expr comma
want_for_map:	  scalar_expr comma
want_for_sort:	  scalar_expr comma

maybe_decl:	  scope_class <commit> variable props['is']
		| variable <commit> props['is']
		| # nada

ENDGRAMMAR

} # END INIT

1;
